@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP                        #-}
 {-# LANGUAGE GADTs                      #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE OverloadedStrings          #-}
@@ -11,9 +12,10 @@ import qualified Blaze.ByteString.Builder           as BB
 import qualified Blaze.ByteString.Builder.Char.Utf8 as BB
 import           Control.Applicative
 import           Control.Monad
+import qualified Control.Monad.Fail                 as F
 import           Data.Attoparsec.ByteString
 import qualified Data.Attoparsec.ByteString         as A
-import qualified Data.Attoparsec.ByteString.Char8   as A ( decimal )
+import qualified Data.Attoparsec.ByteString.Char8   as A (decimal)
 import           Data.Bits
 import           Data.ByteString                    (ByteString)
 import qualified Data.ByteString                    as BS
@@ -25,6 +27,7 @@ import           Data.List                          (delete, intersperse,
 import qualified Data.Map.Strict                    as M
 import           Data.Maybe
 import           Data.Monoid
+import           Data.Semigroup                     (Semigroup)
 import           Data.Ord                           (comparing)
 import           Data.Word
 import           Text.Read                          (readMaybe)
@@ -142,7 +145,7 @@ serializeURI = normalizeURIRef noNormalization
 -- possible, 'serializeURIRef' will be fine. If you intend on
 -- comparing URIs (say for caching purposes), you'll want to use this.
 normalizeURIRef :: URINormalizationOptions -> URIRef a -> Builder
-normalizeURIRef o uri@(URI {..}) = normalizeURI o uri
+normalizeURIRef o uri@(URI {..})       = normalizeURI o uri
 normalizeURIRef o uri@(RelativeRef {}) = normalizeRelativeRef o Nothing uri
 
 
@@ -169,7 +172,8 @@ normalizeRelativeRef o@URINormalizationOptions {..} mScheme RelativeRef {..} =
   authority <> path <> query <> fragment
   where
     path
-      | unoSlashEmptyPath && BS.null rrPath = "/"
+      | unoSlashEmptyPath && BS.null rrPath  = BB.fromByteString "/"
+      | segs == [""] = BB.fromByteString "/"
       | otherwise  = mconcat (intersperse (c8 '/') (map urlEncodePath segs))
     segs = dropSegs (BS.split slash (pathRewrite rrPath))
     pathRewrite
@@ -366,7 +370,7 @@ relativeRefParser' opts = do
   query <- queryParser opts
   frag  <- mFragmentParser
   case frag of
-    Just _ -> endOfInput `orFailWith` MalformedFragment
+    Just _  -> endOfInput `orFailWith` MalformedFragment
     Nothing -> endOfInput `orFailWith` MalformedQuery
   return $ RelativeRef authority path query frag
 
@@ -791,9 +795,18 @@ newtype Parser' e a = Parser' { unParser' :: Parser a}
                              , Alternative
                              , Monad
                              , MonadPlus
+                             , Semigroup
                              , Monoid)
 
 
+instance F.MonadFail (Parser' e) where
+#if MIN_VERSION_attoparsec(0,13,1)
+  fail e = Parser' (F.fail e)
+#else
+  fail e = Parser' (fail e)
+#endif
+
+
 -------------------------------------------------------------------------------
 -- | Use with caution. Catch a parser failing and return Nothing.
 mParse :: Parser' e a -> Parser' e (Maybe a)
@@ -804,7 +817,7 @@ mParse p = option Nothing (Just <$> p)
 -- | If the first parser succeeds, discard the result and use the
 -- second parser (which may fail). If the first parser fails, return
 -- Nothing. This is used to check a benign precondition that indicates
--- the presence of a parsible token, i.e. ? preceeding a query.
+-- the presence of a parsible token, i.e. ? preceding a query.
 thenJust :: Parser' e a -> Parser' e b -> Parser' e (Maybe b)
 thenJust p1 p2 = p1 *> (Just <$> p2) <|> pure Nothing
 
